require "user_dissociation"

class UsersController < ChefObjectController
  # GET /users/:id/groups
  def groups
    # TODO - for efficient listing, chef-server should support /users/<username>/groups
    # For now use the brute force approach
    @groups = server.get_rest("groups")
    # get the group details for members listing.
    @groups = (@groups.keys || []).map { |id|
      server.get_rest("groups/#{id}")}

    # Filter groups for this 'user'
    user = params[:id]
    @groups = @groups.select do |group|
      if USAG.match(group["groupname"])
        false
      else
        !group["users"].index(user).nil?
      end
    end

    respond_to do |format|
      format.json { render :json => @groups }
      format.html { show }
    end
  end

  def update
    render begin
      { json: server_user_root(session_username).put_rest(
        "users/#{params[:id]}", params[:user]) }
    rescue Net::HTTPServerException => e
      { json: parsed_json_for_error(e), status: e.response.code }
    end
  end

  # Same as a user update, but takes current_password, new_password, and
  # new_password_confirmation and validates them.
  def update_password
    # Ensure the current password is present
    if params[:user][:current_password].to_s.blank?
      render json: { error: 'Current password is required.' },
             status: :forbidden and return
    end

    # Ensure the current password is correct
    begin
      server_root.post_rest('authenticate_user',
                           { username: params[:id],
                             password: params[:user][:current_password] })
    rescue Net::HTTPServerException => e
      if e.response.code.to_i == 401
        render json: { error: 'Current password is incorrect.' },
               status: :forbidden and return
      else
        raise e
      end
    end

    # Ensure the new password and confirmation match
    if params[:user][:new_password] != params[:user][:new_password_confirmation]
      render json: { error: 'New password and confirmation do not match.' },
             status: :forbidden and return
    end

    # Modify params to give the server what it wants
    params[:user].delete(:new_password_confirmation)
    params[:user].delete(:current_password)
    params[:user][:username] = params[:id]
    params[:user][:password] = params[:user].delete(:new_password)

    begin
      render json: server_user_root(session_username).put_rest(
        "users/#{params[:id]}", params[:user])

      # Reset password reset time; this will invalidate all older sessions and force
      # them to log back in with the new password
      user_store = ChefManageUserStore.new(Settings.redis.url)
      user_store.password_reset(session[:username])

      # Except this one, since we set a new session time
      session[:session_start_time] = Time.zone.now
    rescue Net::HTTPServerException => e
      if e.response.code.to_i == 400
        render json: parsed_json_for_error(e), status: :forbidden and return
      else
        raise e
      end
    end
  end

  def remove
    ##############################################################################
    # TODO: need to check that session_username has the auth to rm params[:user] #
    ##############################################################################
    # :failure => "not-authorized"
    leave_org = UserDissociation.new(server_root, params[:user], params[:organization_id])

    # if the user leaving this org would make the admins group empty, render 403
    if leave_org.last_user_in_admins_group?
      render :json => {:success => false,
                       :message => "This user leaving the org would leave the admins group empty",
                       :failure => "admins"}, :status => 403

    # if the user leaving this org would make the billing-admins group empty, render 403
    elsif leave_org.last_user_in_billing_admins_group?
      render :json => {:success => false,
                       :message => "This user leaving the org would leave the billing-admins group empty",
                       :failure => "billing-admins"}, :status => 403
    else

      begin
        server.delete_rest("users/#{params[:user]}")
        render :json => {:success => true}, :status => 200
      rescue Net::HTTPServerException => e
        # server enforces a rule where an admin cannot be deleted from an organization - they must
        # first be removed from the admins group. If that happens, it's a 403 with a detailed error message.
        # A side effect here is if this is a permissions error, we'll now be displaying that error as well.
        # This will usually be along the lines of '$user does not have access to update an organization'
        if e.response.code.to_i == 403
          render json: parsed_json_for_error(e), status: :forbidden and return
        else
          raise e
        end
      end
    end
  end

  # PUT /users/:id/accept_invite/:invite_request_id
  # TODO:
  # Super confusing that you reject individual invites via:
  # DELETE organizations/:reject_org/association_requests/:invite_request_id
  #
  # But you accept individaul invites via:
  # PUT /users/:id/association_requests/:invite_request_id
  #
  # But that is how the API does it afaik.
  def accept_invite
    result = server_user_root(session_username).put_rest("users/#{params[:id]}/association_requests/#{params[:invite_request_id]}", {:response => 'accept'})
    render :json => {:result => result}
  end

  protected

  def index_url
    "users"
  end

  def get_objects_details(from, to)
    (@objects[from..to] || []).map {|id| object_for_id(id["user"]["username"]) }
  end
end
